"""
Taylor moment expansion (TME) in JaX.

For math details, please see the docstring of :py:mod:`tme.base_sympy`.

Functions
---------
:py:func:`generator`
    Infinitesimal generator. This is a helper function around :py:func:`generator_power`.
:py:func:`generator_power`
    Iterations/power of infinitesimal generators.
:py:func:`mean_and_cov`
    TME approximation for mean and covariance. In case you just want to compute the mean, use function
    :py:func:`expectation` with argument :code:`phi` fed by an identity function.
:py:func:`expectation`
    TME approximation for any expectation of the form :math:`\mathbb{E}[\phi(X(t + \Delta t)) \mid X(t)]`.

References
----------
See the docstring of :py:mod:`tme.base_sympy`.

Authors
-------
Adrien Corenflos and Zheng Zhao, 2021
"""
# TODO: The logic further down can be improved dramatically if we make diagonal noise specific logic.

try:
    import jax as _
except:
    raise ImportError("By default the library is not packaged with JaX due to the need to support CPU and GPU users. "
                      "In order to use it, follow the instructions on https://github.com/google/jax#installation.")

import math
from math import factorial
from typing import Callable, List, Tuple

import jax.numpy as jnp
from jax import jvp, linearize, vmap

__all__ = ['generator',
           'generator_power',
           'mean_and_cov',
           'expectation']


def generator_power(phi: Callable, drift: Callable, dispersion: Callable, order: int = 1) -> List[Callable]:
    r"""Iterations/power of infinitesimal generator.

    For math details, see the docstring of :py:func:`tme.base_sympy.generator_power`.

    Parameters
    ----------
    phi : Callable (d,) -> (...)
        Target function.
    drift : Callable (d,) -> (d,)
        SDE drift coefficient.
    dispersion : Callable (d,) -> (d, w)
        SDE dispersion coefficient, where `w` stands for the dimension of the Wiener process.
    order : int, optional
        Number of generator iterations. Must be >=0. Default is 1, which corresponds to the standard infinitesimal
        generator.

    Returns
    -------
    List[Callable]
        List of generator functions in ascending power order. Formally, this function returns
        :math:`[\phi, \mathcal{A}\phi, \ldots, \mathcal{A}^p\phi]`, where :code:`p` is the order.
        Each callable function in this list has exactly the same input-output shape
        signature as phi:  (d,) -> (...).

    Notes
    -----
    The implementation is due to Adrien Corenflos. Thank you for contributing this.

    You may also find a naive implementation of infinitesimal generators and their iterations in the test file
    :code:`./test/test_tme_jax.py`.
    """

    def jac_part(z, f):
        # This computes the Jacobian-vector product J[f](z) * a(z)
        return jvp(f, (z,), (drift(z),))[1]

    def hess_prod_1(z, f):
        # This computes the Jacobian-vector product J[f](z) * b(z)
        _, linearized_f = linearize(f, z)
        return vmap(linearized_f, in_axes=1, out_axes=0)(_format_dispersion(dispersion(z)))

    def hess_prod_2(z, f):
        # This computes the double Jacobian-vector product J[z -> J[f](z) * b(z)](z) * b(z).T
        # This is an equivalent, but more efficient way, to computing the Hessian form b(z).T * H[f](z) * b(z)
        # TODO: Verify is linearize is faster than vectorized vjp here.
        _, linearized_f = linearize(lambda zz: hess_prod_1(zz, f), z)
        return vmap(linearized_f, in_axes=0, out_axes=1)(_format_dispersion(dispersion(z)).T)

    def hess_part(z, f):
        # This computes the trace of the matrix product batched along the trailing dimensions of the Hessian.
        return jnp.einsum("ii...", hess_prod_2(z, f))

    gen_power = phi

    list_of_gen_powers = [gen_power]

    for _ in range(order):
        def gen_power(z, f=gen_power):
            return jac_part(z, f) + 0.5 * hess_part(z, f)

        list_of_gen_powers.append(gen_power)

    return list_of_gen_powers


def generator(phi: Callable, drift: Callable, dispersion: Callable) -> Callable:
    r"""Infinitesimal generator for diffusion processes in Ito's SDE constructions.

    .. math::

        (\mathcal{A}\phi)(x) = \sum^d_{i=1} a_i(x)\,\frac{\partial \phi}{\partial x_i}(x)
        + \frac{1}{2}\, \sum^d_{i,j=1} \Gamma_{ij}(x) \, \frac{\partial^2 \phi}{\partial x_i \, \partial x_j}(x),

    where :math:`\phi\colon \mathbb{R}^d \to \mathbb{R}` must be sufficiently smooth function depending on the
    expansion order, and :math:`\Gamma(x) = b(x) \, b(x)^\top`.

    This is a helper function around :py:func:`generator_power`.

    Parameters
    ----------
    phi : Callable (d,) -> (...)
        Target function.
    drift : Callable (d,) -> (d,)
        SDE drift coefficient.
    dispersion : Callable (d,) -> (d, w)
        SDE dispersion coefficient, where `w` stands for the dimension of the Wiener process.

    Returns
    -------
    Callable (...)
        A callable function which carries out :math:`x \mapsto \mathcal{A}\phi`. The output shape of this function
        is the same as :code:`phi`.
    """
    return generator_power(phi, drift, dispersion, 1)[1]


def mean_and_cov(x: jnp.ndarray, dt: float, drift: Callable, dispersion: Callable,
                 order: int = 3) -> Tuple[jnp.ndarray, jnp.ndarray]:
    r"""TME approximation for mean and covariance.

    For math details, see the docstring of :py:func:`tme.base_sympy.mean_and_cov`.

    Parameters
    ----------
    x : jnp.ndarray (d,)
        The state at which the generator is evaluated. (i.e., the :math:`x` in
        :math:`\mathbb{E}[X(t + \Delta t) \mid X(t)=x]` and :math:`\mathrm{Cov}[X(t + \Delta t) \mid X(t)=x]`).
    dt : float
        Time interval.
    drift : Callable (d,) -> (d,)
        SDE drift coefficient.
    dispersion : Callable (d,) -> (d, w)
        SDE dispersion coefficient, where `w` stands for the dimension of the Wiener process.
    order : int, default=3
        Order of TME. Must be >= 1.

    Returns
    -------
    m : jnp.ndarray (d,)
        TME approximation of mean :math:`\mathbb{E}[X(t + \Delta t) \mid X(t)=x]`.
    cov : jnp.ndarray (d, d)
        TME approximation of covariance :math:`\mathrm{Cov}[X(t + \Delta t) \mid X(t)=x]`.

    Notes
    -----
    When `order = 1`, the TME mean and cov approximations are exactly the same with Euler--Maruyama.
    """
    # Give generator powers of phi^I and phi^II then evaluate them all
    list_of_Aphi_i = generator_power(lambda z: z, drift, dispersion, order)
    list_of_Aphi_ii = generator_power(lambda z: jnp.outer(z, z), drift, dispersion, order)

    Aphi_i_powers = [func(x) for func in list_of_Aphi_i]
    Aphi_ii_powers = [func(x) for func in list_of_Aphi_ii]

    # Give the mean approximation
    m = x
    for r in range(1, order + 1):
        m = m + 1 / factorial(r) * Aphi_i_powers[r] * dt ** r

    # Give the cov approximation
    # r = 1
    cov = Aphi_ii_powers[1] - jnp.outer(Aphi_i_powers[0], Aphi_i_powers[1]) \
          - jnp.outer(Aphi_i_powers[1], Aphi_i_powers[0])
    cov = cov * dt

    for r in range(2, order + 1):
        coeff = Aphi_ii_powers[r]
        for k in range(r + 1):
            coeff = coeff - _comb(r, k) * jnp.outer(Aphi_i_powers[k], Aphi_i_powers[r - k])
        cov = cov + 1 / factorial(r) * coeff * dt ** r

    return m, cov


def expectation(phi: Callable, x: jnp.ndarray, dt: float, drift: Callable, dispersion: Callable,
                order: int = 3) -> jnp.ndarray:
    r"""TME approximation of expectation on any target function :math:`\phi`.

    For math details, see the docstring of :py:func:`tme.base_sympy.expectation`.

    Parameters
    ----------
    phi : Callable (d,) -> (...)
        Target function (must be sufficiently smooth depending on the order).
    x : jnp.ndarray (d, )
        The state at which the generator is evaluated (i.e., the :math:`x` in
        :math:`\mathbb{E}[\phi(X(t + \Delta t)) \mid X(t)=x]`).
    dt : float
        Time interval.
    drift : Callable (d,) -> (d,)
        SDE drift coefficient.
    dispersion : Callable (d,) -> (d, w)
        SDE dispersion coefficient, where `w` stands for the dimension of the Wiener process.
    order : int
        Order of TME. Must be >=0. For the relationship between the expansion order and SDE coefficient smoothness, see,
        Zhao (2021).

    Returns
    -------
    jnp.ndarray (...)
        TME approximation of :math:`\mathbb{E}[\phi(X(t + \Delta t)) \mid X(t)]`. The output shape is consistence with
        the input shape of :code:`phi`.
    """
    list_of_Aphi = generator_power(phi, drift, dispersion, order)
    Aphi = phi(x)
    for r in range(1, order + 1):
        Aphi += 1 / factorial(r) * list_of_Aphi[r](x) * dt ** r

    return Aphi


def _comb(n, k):
    try:
        return math.comb(n, k)
    except AttributeError:  # Python version < 3.8 does not have math.comb
        return _manual_comb(n, k)


def _manual_comb(n, k):
    if k > n // 2:
        return _manual_comb(n, n - k)
    if k < 0:
        return 0
    if k == 0:
        return 1
    return _manual_comb(n - 1, k - 1) + _manual_comb(n - 1, k)


def _format_dispersion(bz):
    ndim = jnp.ndim(bz)
    if ndim == 0:
        return jnp.atleast_2d(bz)
    if ndim == 1:
        return jnp.expand_dims(bz, 1)
    if ndim == 2:
        return bz
    else:
        raise ValueError(f"Dispersion coefficient b(z) must have at most 2 dimensions. {ndim} were passed")



r"""
Taylor moment expansion (TME) in SymPy.

TME applies on stochastic differential equations (SDEs) of the form

    .. math::

        d X(t) = a(X(t))dt + b(X(t)) dW(t),

    where :math:`X \colon \mathbb{T} \to \mathbb{R}^d`, and :math:`W\colon \mathbb{T} \to \mathrm{R}^w` is a Wiener
    process having the unit spectral density.
    Functions :math:`a \colon \mathbb{R}^d \to \mathbb{R}^d`
    and :math:`b \colon \mathbb{R}^d \to \mathbb{R}^{d \times w}` stand for the drift and dispersion coefficients,
    respectively. :math:`\mathbb{T} = [t_0, \infty)` stands for a temporal domain.

For more detailed explanations of TME, see, Zhao (2021) or Zhao et al. (2020). Notations used in this doc-site
follows from Zhao (2021).

Functions
---------
:py:func:`mean_and_cov`
    TME approximation for mean and covariance. In case you just want to compute the mean, use function
    :py:func:`expectation` with argument :code:`phi` fed by an identity function.
:py:func:`expectation`
    TME approximation for any expectation of the form :math:`\mathbb{E}[\phi(X(t + \Delta t)) \mid X(t)]`.
:py:func:`generator`
    Infinitesimal generator.
:py:func:`generator_vec`
    Infinitesimal generator extended for vector-valued target function.
:py:func:`generator_mat`
    Infinitesimal generator extended for matrix-valued target function.
:py:func:`generator_power`
    Iterations/power of infinitesimal generators.

References
----------
Zheng Zhao. State-space deep Gaussian processes.
PhD thesis, Aalto University. 2021.

Zheng Zhao, Toni Karvonen, Roland Hostettler, and Simo Särkkä.
Taylor Moment Expansion for Continuous-Discrete Gaussian Filtering.
IEEE Transactions on Automatic Control, 66(9):4460-4467, 2021.

Didier Dacunha-Castelle and Danielle Florens-Zmirou.
Estimation of the coefficients of a diffusion from discrete observations.
Stochastics, 19(4):263–284, 1986.

Danielle Florens-Zmirou. Approximate discrete-time schemes for statistics of diffusion processes.
Statistics, 20(4):547–557, 1989.

Notes
-----
Currently, this symbolc implementation has a problem as the SymPy simplification function
:code:`sympy.simplify` is not working well. See details in the docstring of function :py:func:`mean_and_cov`.

Authors
-------
Zheng Zhao, 2020, zz@zabemon.com, https://zz.zabemon.com
"""
import warnings
from typing import Tuple, List, Union, Callable

import sympy
from sympy import Symbol, Expr, Matrix, MatrixSymbol, factorial, trace, zeros
from sympy import binomial, simplify

__all__ = ['generator',
           'generator_vec',
           'generator_mat',
           'generator_power',
           'mean_and_cov',
           'expectation']


def generator(phi: Expr, x: MatrixSymbol, drift: Matrix, dispersion: Matrix) -> Expr:
    r"""Infinitesimal generator for diffusion processes in Ito's SDE constructions.

    .. math::

        (\mathcal{A}\phi)(x) = \sum^d_{i=1} a_i(x)\,\frac{\partial \phi}{\partial x_i}(x)
        + \frac{1}{2}\, \sum^d_{i,j=1} \Gamma_{ij}(x) \, \frac{\partial^2 \phi}{\partial x_i \, \partial x_j}(x),

    where :math:`\phi\colon \mathbb{R}^d \to \mathbb{R}` must be sufficiently smooth function depending on the
    expansion order, and :math:`\Gamma(x) = b(x) \, b(x)^\top`.

    Parameters
    ----------
    phi : Expr
        Scalar-valued target function.
    x : MatrixSymbol
        Symbolic state vector.
    drift : Matrix
        Symbolic drift coefficient.
    dispersion : Matrix
        Symbolic dispersion coefficient.

    Returns
    -------
    Expr
        Symbolic :math:`(\mathcal{A}\phi)(x)`.
    """
    return Matrix([phi]).jacobian(x).dot(drift) \
           + 0.5 * trace(Matrix([phi]).jacobian(x).jacobian(x) * (dispersion * dispersion.T))


def generator_vec(phi_vec: Matrix, x: MatrixSymbol, drift: Matrix, dispersion: Matrix) -> Matrix:
    r"""Infinitesimal generator for vector-valued :math:`\phi\colon \mathbb{R}^d\to\mathbb{R}^{m}`.

    Parameters
    ----------
    phi_vec : Matrix
        Vector-valued target function.
    x : MatrixSymbol
        Symbolic state vector.
    drift : Matrix
        Symbolic drift coefficient.
    dispersion : Matrix
        Symbolic dispersion coefficient.

    Returns
    -------
    Matrix
        Symbolic :math:`(\mathcal{A}\phi)(x)`.

    Notes
    -----
    Since we are using :code:`sympy.Matrix`, the function :py:func:`generator_mat` can exactly
    replace this function.
    """
    g = zeros(phi_vec.shape[0], phi_vec.shape[1])

    for i, row in enumerate(phi_vec):
        g[i] = generator(row, x, drift, dispersion)
    return g


def generator_mat(phi_mat: Matrix, x: MatrixSymbol, drift: Matrix, dispersion: Matrix) -> Matrix:
    r"""Infinitesimal generator for matrix-valued :math:`\phi\colon \mathbb{R}^d\to\mathbb{R}^{m\times n}`.

    This function exactly corresponds to the operator :math:`\overline{\mathcal{A}}` in Zhao (2021).

    Parameters
    ----------
    phi_mat : Matrix
        Matrix-valued target function.
    x : MatrixSymbol
        Symbolic state vector.
    drift : Matrix
        Symbolic drift coefficient.
    dispersion : Matrix
        Symbolic dispersion coefficient.

    Returns
    -------
    Matrix
        Symbolic :math:`(\mathcal{A}\phi)(x)`.
    """
    g = zeros(phi_mat.shape[0], phi_mat.shape[1])

    for i in range(phi_mat.rows):
        for j in range(phi_mat.cols):
            g[i, j] = generator(phi_mat[i, j], x, drift, dispersion)
    return g


def generator_power(phi: Matrix, x: MatrixSymbol, drift: Matrix, dispersion: Matrix, p: int) -> List[Matrix]:
    r"""Iterations/power of infinitesimal generator for scalar/vector/matrix-valued :math:`\phi`.

    .. math::

        (\mathcal{A}^p\phi)(x),

    where :math:`p` is the number of iterations.

    Parameters
    ----------
    phi : Matrix
        Scalar/vector/Matrix-valued target function.
    x : MatrixSymbol
        Symbolic state vector.
    drift : Matrix
        Symbolic drift coefficient.
    dispersion : Matrix
        Symbolic dispersion coefficient.
    p : int
        Power/number of iterations.

    Returns
    -------
    List[Matrix]
        A list of iterated generators
        :math:`[(\mathcal{A}^0\phi)(x), (\mathcal{A}\phi)(x), \ldots, (\mathcal{A}^p\phi)(x)]`.
    """
    generators = [phi]

    for i in range(1, p + 1):
        phi = generator_mat(phi, x, drift, dispersion)
        generators.append(phi)
    return generators


def mean_and_cov(x: MatrixSymbol, drift: Matrix, dispersion: Matrix, dt: Symbol, order: int = 3,
                 simp: Union[bool, Callable] = True) -> Tuple[Matrix, Matrix]:
    r"""TME approximation for mean and covariance.

    Formally, this function approximates

    .. math::

        \mathbb{E}[X(t + \Delta t) \mid X(t)],

        \mathrm{Cov}[X(t + \Delta t) \mid X(t)].

    See, Zhao (2021, Lemma 3.4) for details.

    Parameters
    ----------
    x : MatrixSymbol
        Symbolic state vector.
    drift : Matrix
        Symbolic drift coefficient.
    dispersion : Matrix
        Symbolic dispersion coefficient.
    dt : Symbol
        Symbolic time interval. You can create one, for example, by :code:`dt=sympy.Symbol('dt', positive=True)`.
    order : int, default=3
        Order of TME.
    simp : bool or Callable, default=True
        Set :code:`True` to simplify the results by calling the default :code:`sympy.simplify` method.
        You can also use your own simplification method by feeding it a callable function.

        .. warning::

            The method :code:`sympy.simplify` can be unnecessarily slow when the system is complicated,
            see `this page <https://docs.sympy.org/latest/tutorial/simplification.html>`_ for details.

    Returns
    -------
    m : Matrix
        TME approximation of mean :math:`\mathbb{E}[X(t + \Delta t) \mid X(t)]`.
    cov : Matrix
        TME approximation of covariance :math:`\mathrm{Cov}[X(t + \Delta t) \mid X(t)]`.
    """
    if dt is None:
        dt = Symbol('dt', positive=True)

    dim_x = x.shape[0]

    # Give mean estimate
    phi = Matrix([x])
    m = Matrix([x])
    for r in range(1, order + 1):
        phi = generator_vec(phi, x, drift, dispersion)
        m = m + 1 / factorial(r) * phi * dt ** r

    # Give covariance estimate
    # Precompute powers of generator
    Ax = generator_power(Matrix([x]), x, drift, dispersion, order)
    Axx = generator_power(x * x.T, x, drift, dispersion, order)

    cov = sympy.zeros(dim_x, dim_x)
    for r in range(1, order + 1):
        coeff = Axx[r]
        for s in range(r + 1):
            coeff = coeff - binomial(r, s) * Ax[s] * Ax[r - s].T
        cov = cov + 1 / factorial(r) * coeff * dt ** r

    if callable(simp):
        return simp(m), simp(cov)
    if simp:
        warn_simp()
        return simplify(m), simplify(cov)
    return m, cov


def expectation(phi: Matrix, x: MatrixSymbol, drift: Matrix, dispersion: Matrix, dt: Symbol = None,
                order: int = 3, simp: Union[bool, Callable] = True) -> Matrix:
    r"""TME approximation of expectation on target function :math:`\phi`.

    Formally, this function approximates

    .. math::

        \mathbb{E}[\phi(X(t+\Delta t)) \mid X(t)],

    where :math:`\phi \colon \mathbb{R}^d \to \mathbb{R}^{m\times n}` is any smooth enough function of interest.

    Parameters
    ----------
    phi : Matrix
        Target function.
    x : MatrixSymbol
        Symbolic state vector.
    drift : Matrix
        Symbolic drift coefficient.
    dispersion : Matrix
        Symbolic dispersion coefficient.
    dt : Symbol, optional
        Symbolic time interval. If this is not specified by the user, then the function will create one.
    order : int, default=3
        Order of TME.
    simp : bool or Callable, default=True
        Set True to simplify the results by calling the default :code:`sympy.simplify` method. You can also give
        your own simplification method as a callable function input.

    Returns
    -------
    Matrix
        TME approximation of :math:`\mathbb{E}[\phi(X(t + \Delta t)) \mid X(t)]`.
    """
    if dt is None:
        dt = Symbol('dt', positive=True)

    # Precompute generator powers
    Ar = generator_power(phi, x, drift, dispersion, order)

    expec = sympy.zeros(*phi.shape)
    for r in range(order + 1):
        expec = expec + 1 / factorial(r) * Ar[r] * dt ** r

    if callable(simp):
        return simp(expec)
    if simp:
        warn_simp()
        return simplify(expec)
    return expec


def warn_simp():
    warnings.warn('The simplification method sympy.simplify may be slow in some cases.')



import unittest

import jax.numpy as jnp
import numpy as np
import numpy.testing as npt
import sympy as sp
import tme.base_jax as tme_jax
import tme.base_sympy as tme_sp
from jax import hessian, jit, jacfwd
from jax.config import config


def setUpModule():
    config.update("jax_enable_x64", True)


def phi_sym(x):
    return sp.Matrix([[x[0] * x[1], x[1] ** 3]])


def phi_jax(x):
    return jnp.array([x[0] * x[1], x[1] ** 3])


def phi_jax_2d(x):
    return jnp.array([[x[0] * x[1], x[1] ** 3]])


def generator(phi, x, a, b):
    bb = b(x)
    return jacfwd(phi)(x) @ a(x) + 0.5 * jnp.trace(hessian(phi)(x) @ (bb @ bb.T), axis1=-2, axis2=-1)


def generator_power_naive(phi, a, b, order: int):
    list_of_gen_powers = [phi]

    gen_power = phi

    for _ in range(order):
        def gen_power(z, f=gen_power):
            return generator(f, z, a, b)

        list_of_gen_powers.append(gen_power)

    return list_of_gen_powers


class TestJaxVsSymPy(unittest.TestCase):
    """
    Test against SymPy results which should give true results.
    """

    def setUp(self) -> None:
        np.random.seed(111)

        self.dim_x = 2
        self.dim_w = 3

        self.sym_x = sp.MatrixSymbol('x', self.dim_x, 1)
        self.sym_dt = sp.Symbol('dt', positive=True)

        self.a = sp.Matrix([self.sym_x[0] ** 2 + self.sym_x[1],
                            sp.sin(self.sym_x[1])])

        self.b = sp.Matrix([[1., 2., 3.],
                            [1., 1., 1.]])

        self.order = 4

    def gen_Ap_sympy(self):
        return tme_sp.generator_power(phi_sym(self.sym_x), self.sym_x, self.a, self.b, self.order)

    @staticmethod
    def a_jax(z):
        return jnp.array([z[0] ** 2 + z[1],
                          jnp.sin(z[1])])

    @staticmethod
    def b_jax(z):
        return jnp.array([[1., 2., 3.],
                          [1., 1., 1.]])

    def test_power(self):
        """Test power of generator (both generator_power() and generator_power_naive()).
        """
        x = 0.1 * np.random.randn(self.dim_x)

        list_of_Ap_sympy = self.gen_Ap_sympy()
        list_of_Ap_jax = tme_jax.generator_power(phi_jax, self.a_jax, self.b_jax, self.order)
        list_of_Ap_jax_naive = generator_power_naive(phi_jax, self.a_jax, self.b_jax, self.order)

        for Ap_sympy, Ap_jax, Ap_jax_naive in zip(list_of_Ap_sympy, list_of_Ap_jax, list_of_Ap_jax_naive):
            Ap_func_sympy = sp.lambdify([self.sym_x], Ap_sympy, 'numpy')
            Ap_result_sympy = Ap_func_sympy(x.reshape(self.dim_x, 1))

            Ap_result_jax = Ap_jax(jnp.array(x))
            Ap_result_jax_naive = Ap_jax_naive(jnp.array(x))

            npt.assert_allclose(Ap_result_jax, np.squeeze(Ap_result_sympy))
            npt.assert_allclose(Ap_result_jax, Ap_result_jax_naive)

    def test_mean_and_cov(self):
        """Test mean and cov approximations.
        """
        x = 0.1 * np.random.randn(self.dim_x)
        dt = 0.01

        for order in [2, 3, 4]:
            m_sympy, cov_sympy = tme_sp.mean_and_cov(self.sym_x, self.a, self.b, self.sym_dt, order, simp=True)
            m_sympy_func = sp.lambdify([self.sym_x, self.sym_dt], m_sympy, 'numpy')
            cov_sympy_func = sp.lambdify([self.sym_x, self.sym_dt], cov_sympy, 'numpy')

            m_result_sympy = m_sympy_func(x.reshape(self.dim_x, 1), dt)
            cov_result_sympy = cov_sympy_func(x.reshape(self.dim_x, 1), dt)

            @jit
            def jitted_mcov(z):
                return tme_jax.mean_and_cov(z, dt, self.a_jax, self.b_jax, order)

            m_result_jax, cov_result_jax = jitted_mcov(x)

            npt.assert_allclose(m_result_jax, np.squeeze(m_result_sympy))
            npt.assert_allclose(cov_result_jax, np.squeeze(cov_result_sympy))

    def test_expectation(self):
        """Test expectation computations.
        """
        x = 0.1 * np.random.randn(self.dim_x)
        dt = 0.01

        for order in [2, 3, 4]:
            expec_sympy = tme_sp.expectation(phi_sym(self.sym_x), self.sym_x, self.a, self.b, self.sym_dt,
                                             order, simp=False)
            expec_sympy_func = sp.lambdify([self.sym_x, self.sym_dt], expec_sympy, 'numpy')

            expec_result_sympy = expec_sympy_func(x.reshape(self.dim_x, 1), dt)

            @jit
            def jitted_expec(z):
                return tme_jax.expectation(phi_jax, z, dt, self.a_jax, self.b_jax, order)

            expec_result_jax = jitted_expec(x)

            npt.assert_allclose(expec_result_jax, np.squeeze(expec_result_sympy))


class TestvsEulerMaruyama(unittest.TestCase):
    """
    TME with order 1 should be consistent with Euler--Maruyama on approximating mean
    and covariance
    """

    def test_vs_em(self):
        sigma = 10.
        rho = 28.
        beta = 8 / 3

        def drift(u):
            return jnp.array([sigma * (u[1] - u[0]),
                              u[0] * (rho - u[2]) - u[1],
                              u[0] * u[1] - beta * u[2]])

        bb = 0.15 * jnp.eye(3)

        def dispersion(u):
            return bb

        @jit
        def tme_m_cov(u, dt):
            return tme_jax.mean_and_cov(x=u, dt=dt, drift=drift, dispersion=dispersion, order=1)

        @jit
        def em_m_cov(u, dt):
            return u + drift(u) * dt, dispersion(u) @ dispersion(u).T * dt

        x = jnp.array(np.random.randn(3))
        dt = 1.

        tme_m, tme_cov = tme_m_cov(x, dt)
        em_m, em_cov = em_m_cov(x, dt)

        npt.assert_allclose(tme_m, em_m, atol=1e-12)
        npt.assert_allclose(tme_cov, em_cov, atol=1e-12)


class TestAgainstLinearSDE():
    """Test on linear SDEs which have explict mean and cov solutions.
    """
    pass



"""
Test TME disc
"""
import unittest

import tme.base_sympy as tme
from sympy import Symbol, MatrixSymbol, Matrix, sin, cos, eye, exp, simplify, tanh


class TestGenerator(unittest.TestCase):
    """
    Test generators on scale, vec, and matrix-valued phi functions
    """

    def setUp(self) -> None:
        self.dim_x = 3
        self.dim_w = 3

        self.x = MatrixSymbol('x', self.dim_x, 1)

        self.a = Matrix([self.x[0] ** 2 + self.x[1],
                         self.x[0] * self.x[1],
                         sin(self.x[2])])

        self.b = Matrix([[self.x[0], 0, 0],
                         [0, self.x[1], self.x[1]],
                         [0, 0, self.x[2]]])

    def test_scalar(self) -> None:
        """
        Test generator()
        """
        x = self.x

        phi = x[1] * x[2]
        true_result = x[0] * x[1] * x[2] + x[1] * sin(x[2]) + x[1] * x[2]

        assert (simplify(tme.generator(phi, x, self.a, self.b) - true_result) == 0)

    def test_vector(self) -> None:
        """
        Test generator_vec()
        """
        x = self.x

        phi = Matrix([[x[1] * x[2]],
                      [exp(x[0]) + sin(x[1] + x[2])]])
        true_result = Matrix([[x[0] * x[1] * x[2] + x[1] * sin(x[2]) + x[1] * x[2]],
                              [
                                  exp(x[0]) * (x[0] ** 2 + x[1]) + cos(x[1] + x[2]) * (x[0] * x[1] + sin(x[2]))
                                  + 0.5 * (
                                          exp(x[0]) * x[0] ** 2 - sin(x[1] + x[2]) * (
                                          2 * x[1] ** 2 + x[2] ** 2 + 2 * x[1] * x[2])
                                  )
                              ]])

        assert (simplify(tme.generator_vec(phi, x, self.a, self.b) - true_result) == Matrix([[0], [0]]))

    def test_matrix(self) -> None:
        """
        Test generator_mat()
        """
        x = self.x

        phi = Matrix([[x[1], x[2]],
                      [x[0], sin(x[1])]])
        true_result = Matrix([[x[0] * x[1], sin(x[2])],
                              [x[0] ** 2 + x[1], cos(x[1]) * x[0] * x[1] - sin(x[1]) * x[1] ** 2]])

        assert (simplify(tme.generator_mat(phi, x, self.a, self.b) - true_result) == Matrix([[0, 0], [0, 0]]))

    def test_power(self) -> None:
        """
        Test generator_power
        """
        x = self.x

        order = 2

        phi = Matrix([[x[0]],
                      [x[2]]])

        true_result = [phi, Matrix([[x[0] ** 2 + x[1]],
                                    [sin(x[2])]]), Matrix([[2 * x[0] * (x[0] ** 2 + x[1]) + x[0] * x[1] + x[0] ** 2],
                                                           [cos(x[2]) * sin(x[2]) - 0.5 * sin(x[2]) * x[2] ** 2]])]
        results = tme.generator_power(phi, x, self.a, self.b, order)

        for i, j in zip(true_result, results):
            assert (simplify(i - j) == Matrix([[0], [0]]))


class TestBenes(unittest.TestCase):
    """
    Test TME on a Benese model
    """

    def setUp(self) -> None:
        self.x = MatrixSymbol('x', 1, 1)

        # IMPORTANT: In scalar case, you have to use x[0] instead of x
        self.a = Matrix([tanh(self.x[0])])

        self.b = eye(1)

        self.Q = eye(1)

        self.dt = Symbol('dt', positive=True)

    def test_tme(self) -> None:
        x = self.x
        dt = self.dt

        m, sigma = tme.mean_and_cov(self.x, self.a, self.b, self.dt, order=3, simp=True)

        # Note that there is a typo in the paper.
        # \\phi_{x, 2} after equation 23 should multiply by 2.
        true_a = Matrix([x[0] + tanh(x[0]) * dt])
        true_sigma = Matrix([dt + (1 - tanh(x[0]) ** 2) * dt ** 2])

        assert (simplify(m - true_a) == Matrix([0]))
        assert (simplify(sigma - true_sigma) == Matrix([0]))


if __name__ == '__main__':
    unittest.main()



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name="tme",
    version="0.1.4",
    author="Zheng Zhao",
    author_email="zz@zabemon.com",
    description="Taylor moment expansion in Python.",
    keywords=["stochastic differential equations",
              "statistics"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPL v3 or later",
    url="https://github.com/zgbkdlm/tme",
    download_url="https://github.com/zgbkdlm/tme",
    packages=["tme"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics"
    ],
    python_requires='>=3.8',
    install_requires=requirements
)
